FFmpeg
Loading...
Searching...
No Matches
vf_dblur.c
Go to the documentation of this file.
1/*
2 * Copyright (c) 2020 Paul B Mahol
3 *
4 * This file is part of FFmpeg.
5 *
6 * FFmpeg is free software; you can redistribute it and/or
7 * modify it under the terms of the GNU Lesser General Public
8 * License as published by the Free Software Foundation; either
9 * version 2.1 of the License, or (at your option) any later version.
10 *
11 * FFmpeg is distributed in the hope that it will be useful,
12 * but WITHOUT ANY WARRANTY; without even the implied warranty of
13 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
14 * Lesser General Public License for more details.
15 *
16 * You should have received a copy of the GNU Lesser General Public
17 * License along with FFmpeg; if not, write to the Free Software
18 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
19 */
20
21#include "libavutil/imgutils.h"
22#include "libavutil/mem.h"
23#include "libavutil/opt.h"
24#include "libavutil/pixdesc.h"
25#include "avfilter.h"
26#include "filters.h"
27#include "video.h"
28
29typedef struct DBlurContext {
30 const AVClass *class;
31
32 float angle;
33 float radius;
34 int planes;
35
36 float b0, b1, q, c, R3;
37
38 int depth;
39 int planewidth[4];
41 float *buffer;
44
45#define OFFSET(x) offsetof(DBlurContext, x)
46#define FLAGS AV_OPT_FLAG_VIDEO_PARAM|AV_OPT_FLAG_FILTERING_PARAM|AV_OPT_FLAG_RUNTIME_PARAM
47
48static const AVOption dblur_options[] = {
49 { "angle", "set angle", OFFSET(angle), AV_OPT_TYPE_FLOAT, {.dbl=45}, 0.0, 360, FLAGS },
50 { "radius", "set radius", OFFSET(radius), AV_OPT_TYPE_FLOAT, {.dbl=5}, 0, 8192, FLAGS },
51 { "planes", "set planes to filter", OFFSET(planes), AV_OPT_TYPE_INT, {.i64=0xF}, 0, 0xF, FLAGS },
52 { NULL }
53};
54
56
57#define f(n, m) (dst[(n) * width + (m)])
58
60{
61 DBlurContext *s = ctx->priv;
62 const float b0 = s->b0;
63 const float b1 = s->b1;
64 const float q = s->q;
65 const float c = s->c;
66 float *dst = s->buffer;
67 float g;
68
69 if (s->R3 > 0) {
70 for (int y = 1; y < height; y++) {
71 g = q * f(y, 0) + c * f(y, 0);
72 for (int x = 0; x < width; x++) {
73 f(y, x) = b0 * f(y, x) + b1 * f(y - 1, x) + g;
74 g = q * f(y, x) + c * f(y - 1, x);
75 }
76 }
77
78 for (int y = height - 2; y >= 0; y--) {
79 g = q * f(y, width - 1) + c * f(y, width - 1);
80 for (int x = width - 1; x >= 0; x--) {
81 f(y, x) = b0 * f(y, x) + b1 * f(y + 1, x) + g;
82 g = q * f(y, x) + c * f(y + 1, x);
83 }
84 }
85 } else {
86 for (int y = 1; y < height; y++) {
87 g = q * f(y, width - 1) + c * f(y, width - 1);
88 for (int x = width - 1; x >= 0; x--) {
89 f(y, x) = b0 * f(y, x) + b1 * f(y - 1, x) + g;
90 g = q * f(y, x) + c * f(y - 1, x);
91 }
92 }
93
94 for (int y = height - 2; y >= 0; y--) {
95 g = q * f(y, 0) + c * f(y, 0);
96 for (int x = 0; x < width; x++) {
97 f(y, x) = b0 * f(y, x) + b1 * f(y + 1, x) + g;
98 g = q * f(y, x) + c * f(y + 1, x);
99 }
100 }
101 }
102
103 return 0;
104}
105
106static void diriir2d(AVFilterContext *ctx, int plane)
107{
108 DBlurContext *s = ctx->priv;
109 const int width = s->planewidth[plane];
110 const int height = s->planeheight[plane];
111
113}
114
137
139{
140 DBlurContext *s = ctx->priv;
141
142 av_freep(&s->buffer);
143}
144
145static int config_input(AVFilterLink *inlink)
146{
148 DBlurContext *s = inlink->dst->priv;
149
150 uninit(inlink->dst);
151
152 s->depth = desc->comp[0].depth;
153 s->planewidth[1] = s->planewidth[2] = AV_CEIL_RSHIFT(inlink->w, desc->log2_chroma_w);
154 s->planewidth[0] = s->planewidth[3] = inlink->w;
155 s->planeheight[1] = s->planeheight[2] = AV_CEIL_RSHIFT(inlink->h, desc->log2_chroma_h);
156 s->planeheight[0] = s->planeheight[3] = inlink->h;
157
158 s->nb_planes = av_pix_fmt_count_planes(inlink->format);
159
160 s->buffer = av_malloc_array(FFALIGN(inlink->w, 16), FFALIGN(inlink->h, 16) * sizeof(*s->buffer));
161 if (!s->buffer)
162 return AVERROR(ENOMEM);
163
164 return 0;
165}
166
167static void set_params(DBlurContext *s, float angle, float r)
168{
169 float mu, nu, R1, R2, w1, w2;
170 float a0, a1, a2, a3;
171
172 angle = angle * M_PI / 180.f;
173
174 mu = cosf(angle);
175 nu = sinf(angle);
176 R1 = (mu * r) * (mu * r);
177 R2 = (nu * r) * (nu * r);
178 s->R3 = mu * nu * r * r;
179 w1 = sqrtf(0.25f + R1);
180 w2 = sqrtf(0.25f + R2);
181 a0 = (w1 + 0.5f) * (w2 + 0.5f) - fabsf(s->R3);
182 a1 = 0.5f + w2 - a0;
183 a2 = 0.5f + w1 - a0;
184 a3 = a0 - w1 - w2;
185 s->b0 = 1.f / a0;
186 s->b1 = -a2 / a0;
187 s->q = -a1 / a0;
188 s->c = -a3 / a0;
189}
190
191static int filter_frame(AVFilterLink *inlink, AVFrame *in)
192{
193 AVFilterContext *ctx = inlink->dst;
194 DBlurContext *s = ctx->priv;
195 AVFilterLink *outlink = ctx->outputs[0];
196 AVFrame *out;
197 int plane;
198
199 set_params(s, s->angle, s->radius);
200
201 if (av_frame_is_writable(in)) {
202 out = in;
203 } else {
204 out = ff_get_video_buffer(outlink, outlink->w, outlink->h);
205 if (!out) {
206 av_frame_free(&in);
207 return AVERROR(ENOMEM);
208 }
210 }
211
212 for (plane = 0; plane < s->nb_planes; plane++) {
213 const int height = s->planeheight[plane];
214 const int width = s->planewidth[plane];
215 float *bptr = s->buffer;
216 const uint8_t *src = in->data[plane];
217 const uint16_t *src16 = (const uint16_t *)in->data[plane];
218 const float *src32 = (const float *)in->data[plane];
219 uint8_t *dst = out->data[plane];
220 uint16_t *dst16 = (uint16_t *)out->data[plane];
221 float *dst32 = (float *)out->data[plane];
222 int y, x;
223
224 if (!(s->planes & (1 << plane))) {
225 if (out != in)
226 av_image_copy_plane(out->data[plane], out->linesize[plane],
227 in->data[plane], in->linesize[plane],
228 width * ((s->depth + 7) / 8), height);
229 continue;
230 }
231
232 if (s->depth == 8) {
233 for (y = 0; y < height; y++) {
234 for (x = 0; x < width; x++) {
235 bptr[x] = src[x];
236 }
237 bptr += width;
238 src += in->linesize[plane];
239 }
240 } else if (s->depth <= 16) {
241 for (y = 0; y < height; y++) {
242 for (x = 0; x < width; x++) {
243 bptr[x] = src16[x];
244 }
245 bptr += width;
246 src16 += in->linesize[plane] / 2;
247 }
248 } else {
249 for (y = 0; y < height; y++) {
250 for (x = 0; x < width; x++) {
251 memcpy(bptr, src32, width * sizeof(float));
252 }
253 bptr += width;
254 src32 += in->linesize[plane] / 4;
255 }
256 }
257
258 diriir2d(ctx, plane);
259
260 bptr = s->buffer;
261 if (s->depth == 8) {
262 for (y = 0; y < height; y++) {
263 for (x = 0; x < width; x++) {
264 dst[x] = av_clip_uint8(lrintf(bptr[x]));
265 }
266 bptr += width;
267 dst += out->linesize[plane];
268 }
269 } else if (s->depth <= 16) {
270 for (y = 0; y < height; y++) {
271 for (x = 0; x < width; x++) {
272 dst16[x] = av_clip_uintp2_c(lrintf(bptr[x]), s->depth);
273 }
274 bptr += width;
275 dst16 += out->linesize[plane] / 2;
276 }
277 } else {
278 for (y = 0; y < height; y++) {
279 for (x = 0; x < width; x++) {
280 memcpy(dst32, bptr, width * sizeof(float));
281 }
282 bptr += width;
283 dst32 += out->linesize[plane] / 4;
284 }
285 }
286 }
287
288 if (out != in)
289 av_frame_free(&in);
290 return ff_filter_frame(outlink, out);
291}
292
293static const AVFilterPad dblur_inputs[] = {
294 {
295 .name = "default",
296 .type = AVMEDIA_TYPE_VIDEO,
297 .config_props = config_input,
298 .filter_frame = filter_frame,
299 },
300};
301
303 .p.name = "dblur",
304 .p.description = NULL_IF_CONFIG_SMALL("Apply Directional Blur filter."),
305 .p.priv_class = &dblur_class,
307 .priv_size = sizeof(DBlurContext),
308 .uninit = uninit,
312 .process_command = ff_filter_process_command,
313};
uint8_t ptrdiff_t const uint8_t ptrdiff_t int intptr_t intptr_t int int16_t * dst
Definition dsp.h:87
static int config_input(AVFilterLink *inlink)
const FFFilter ff_vf_dblur
Definition vf_dblur.c:302
int ff_filter_frame(AVFilterLink *link, AVFrame *frame)
Send a frame of data to the next filter.
Definition avfilter.c:1068
int ff_filter_process_command(AVFilterContext *ctx, const char *cmd, const char *arg, char *res, int res_len, int flags)
Generic processing of user supplied commands that are set in the same way as the filter options.
Definition avfilter.c:906
Main libavfilter public API header.
#define s(width, name)
Definition cbs_vp9.c:198
#define FLAGS
Definition cmdutils.c:598
#define AV_CEIL_RSHIFT(a, b)
Definition common.h:60
static av_always_inline av_const unsigned av_clip_uintp2_c(int a, int p)
Clip a signed integer to an unsigned power of two range.
Definition common.h:280
#define av_clip_uint8
Definition common.h:106
#define NULL
Definition coverity.c:32
static __device__ float sqrtf(float a)
static __device__ float fabsf(float a)
static int filter_frame(DBEDecodeContext *s, AVFrame *frame)
Definition dolby_e.c:1067
@ AV_OPT_TYPE_INT
Underlying C type is int.
Definition opt.h:258
@ AV_OPT_TYPE_FLOAT
Underlying C type is float.
Definition opt.h:270
#define AVFILTER_FLAG_SUPPORT_TIMELINE_GENERIC
Some filters support a generic "enable" expression option that can be used to enable or disable a fil...
Definition avfilter.h:196
#define AVERROR(e)
Definition error.h:45
int av_frame_is_writable(AVFrame *frame)
Check if the frame data is writable.
Definition frame.c:535
void av_frame_free(AVFrame **frame)
Free the frame and any dynamically allocated objects in it, e.g.
Definition frame.c:64
int av_frame_copy_props(AVFrame *dst, const AVFrame *src)
Copy only "metadata" fields from src to dst.
Definition frame.c:599
@ AVMEDIA_TYPE_VIDEO
Definition avutil.h:200
void av_image_copy_plane(uint8_t *dst, int dst_linesize, const uint8_t *src, int src_linesize, int bytewidth, int height)
Copy image plane from src to dst.
Definition imgutils.c:374
misc image utilities
#define r
Definition input.c:42
static av_cold void uninit(AVBitStreamFilterContext *ctx)
#define FILTER_INPUTS(array)
Definition filters.h:264
#define FILTER_OUTPUTS(array)
Definition filters.h:265
#define FILTER_PIXFMTS_ARRAY(array)
Definition filters.h:244
#define AVFILTER_DEFINE_CLASS(fname)
Definition filters.h:478
#define av_cold
Definition attributes.h:117
#define NULL_IF_CONFIG_SMALL(x)
Return NULL if CONFIG_SMALL is true, otherwise the argument without modification.
Definition internal.h:88
static enum AVPixelFormat pix_fmts[]
Definition libkvazaar.c:296
#define sinf(x)
Definition libm.h:421
#define cosf(x)
Definition libm.h:80
#define lrintf(x)
Definition libm_mips.h:72
const char * desc
Definition libsvtav1.c:83
static const struct @257111027162314367033347246032313251342043035002 planes[]
#define FFALIGN(x, a)
Definition macros.h:78
#define M_PI
Definition mathematics.h:67
Memory handling functions.
AVOptions.
int av_pix_fmt_count_planes(enum AVPixelFormat pix_fmt)
Definition pixdesc.c:3500
const AVPixFmtDescriptor * av_pix_fmt_desc_get(enum AVPixelFormat pix_fmt)
Definition pixdesc.c:3460
#define AV_PIX_FMT_GBRAP12
Definition pixfmt.h:569
#define AV_PIX_FMT_YUV420P16
Definition pixfmt.h:556
#define AV_PIX_FMT_GBRPF32
Definition pixfmt.h:584
#define AV_PIX_FMT_YUV444P12
Definition pixfmt.h:552
#define AV_PIX_FMT_YUV444P9
Definition pixfmt.h:544
#define AV_PIX_FMT_YUV420P10
Definition pixfmt.h:545
#define AV_PIX_FMT_YUV440P12
Definition pixfmt.h:551
#define AV_PIX_FMT_GRAY9
Definition pixfmt.h:524
#define AV_PIX_FMT_GBRAP16
Definition pixfmt.h:571
#define AV_PIX_FMT_GBRP9
Definition pixfmt.h:563
#define AV_PIX_FMT_YUV422P9
Definition pixfmt.h:543
#define AV_PIX_FMT_YUVA444P10
Definition pixfmt.h:598
#define AV_PIX_FMT_YUVA420P16
Definition pixfmt.h:601
#define AV_PIX_FMT_YUV420P12
Definition pixfmt.h:549
#define AV_PIX_FMT_YUVA420P10
Definition pixfmt.h:596
#define AV_PIX_FMT_YUVA422P9
Definition pixfmt.h:594
#define AV_PIX_FMT_YUV422P12
Definition pixfmt.h:550
#define AV_PIX_FMT_GBRP10
Definition pixfmt.h:564
#define AV_PIX_FMT_YUV422P10
Definition pixfmt.h:546
#define AV_PIX_FMT_GRAY12
Definition pixfmt.h:526
#define AV_PIX_FMT_GBRP12
Definition pixfmt.h:565
#define AV_PIX_FMT_YUV420P9
Definition pixfmt.h:542
#define AV_PIX_FMT_YUVA420P9
Definition pixfmt.h:593
#define AV_PIX_FMT_YUVA422P10
Definition pixfmt.h:597
#define AV_PIX_FMT_GRAYF32
Definition pixfmt.h:588
#define AV_PIX_FMT_YUV420P14
Definition pixfmt.h:553
AVPixelFormat
Pixel format.
Definition pixfmt.h:71
@ AV_PIX_FMT_NONE
Definition pixfmt.h:72
@ AV_PIX_FMT_YUV420P
planar YUV 4:2:0, 12bpp, (1 Cr & Cb sample per 2x2 Y samples)
Definition pixfmt.h:73
@ AV_PIX_FMT_YUV440P
planar YUV 4:4:0 (1 Cr & Cb sample per 1x2 Y samples)
Definition pixfmt.h:106
@ AV_PIX_FMT_YUV422P
planar YUV 4:2:2, 16bpp, (1 Cr & Cb sample per 2x1 Y samples)
Definition pixfmt.h:77
@ AV_PIX_FMT_GRAY8
Y , 8bpp.
Definition pixfmt.h:81
@ AV_PIX_FMT_YUVA420P
planar YUV 4:2:0, 20bpp, (1 Cr & Cb sample per 2x2 Y & A samples)
Definition pixfmt.h:108
@ AV_PIX_FMT_YUVJ440P
planar YUV 4:4:0 full scale (JPEG), deprecated in favor of AV_PIX_FMT_YUV440P and setting color_range
Definition pixfmt.h:107
@ AV_PIX_FMT_YUV410P
planar YUV 4:1:0, 9bpp, (1 Cr & Cb sample per 4x4 Y samples)
Definition pixfmt.h:79
@ AV_PIX_FMT_YUV411P
planar YUV 4:1:1, 12bpp, (1 Cr & Cb sample per 4x1 Y samples)
Definition pixfmt.h:80
@ AV_PIX_FMT_YUV444P
planar YUV 4:4:4, 24bpp, (1 Cr & Cb sample per 1x1 Y samples)
Definition pixfmt.h:78
@ AV_PIX_FMT_YUVA444P
planar YUV 4:4:4 32bpp, (1 Cr & Cb sample per 1x1 Y & A samples)
Definition pixfmt.h:174
@ AV_PIX_FMT_YUVJ411P
planar YUV 4:1:1, 12bpp, (1 Cr & Cb sample per 4x1 Y samples) full scale (JPEG), deprecated in favor ...
Definition pixfmt.h:283
@ AV_PIX_FMT_GBRAP
planar GBRA 4:4:4:4 32bpp
Definition pixfmt.h:212
@ AV_PIX_FMT_YUVJ422P
planar YUV 4:2:2, 16bpp, full scale (JPEG), deprecated in favor of AV_PIX_FMT_YUV422P and setting col...
Definition pixfmt.h:86
@ AV_PIX_FMT_YUVA422P
planar YUV 4:2:2 24bpp, (1 Cr & Cb sample per 2x1 Y & A samples)
Definition pixfmt.h:173
@ AV_PIX_FMT_GBRP
planar GBR 4:4:4 24bpp
Definition pixfmt.h:165
@ AV_PIX_FMT_YUVJ444P
planar YUV 4:4:4, 24bpp, full scale (JPEG), deprecated in favor of AV_PIX_FMT_YUV444P and setting col...
Definition pixfmt.h:87
@ AV_PIX_FMT_YUVJ420P
planar YUV 4:2:0, 12bpp, full scale (JPEG), deprecated in favor of AV_PIX_FMT_YUV420P and setting col...
Definition pixfmt.h:85
#define AV_PIX_FMT_YUVA422P12
Definition pixfmt.h:599
#define AV_PIX_FMT_YUV422P14
Definition pixfmt.h:554
#define AV_PIX_FMT_GRAY10
Definition pixfmt.h:525
#define AV_PIX_FMT_GRAY14
Definition pixfmt.h:527
#define AV_PIX_FMT_YUV422P16
Definition pixfmt.h:557
#define AV_PIX_FMT_GRAY16
Definition pixfmt.h:528
#define AV_PIX_FMT_GBRAP10
Definition pixfmt.h:568
#define AV_PIX_FMT_YUVA444P16
Definition pixfmt.h:603
#define AV_PIX_FMT_YUVA422P16
Definition pixfmt.h:602
#define AV_PIX_FMT_GBRP16
Definition pixfmt.h:567
#define AV_PIX_FMT_YUV444P14
Definition pixfmt.h:555
#define AV_PIX_FMT_YUVA444P9
Definition pixfmt.h:595
#define AV_PIX_FMT_GBRP14
Definition pixfmt.h:566
#define AV_PIX_FMT_YUVA444P12
Definition pixfmt.h:600
#define AV_PIX_FMT_GBRAPF32
Definition pixfmt.h:585
#define AV_PIX_FMT_YUV444P16
Definition pixfmt.h:558
#define AV_PIX_FMT_YUV444P10
Definition pixfmt.h:548
#define R2
#define R1
Describe the class of an AVClass context structure.
Definition log.h:76
An instance of a filter.
Definition avfilter.h:273
void * priv
private data for use by the filter
Definition avfilter.h:288
A filter pad used for either input or output.
Definition filters.h:40
This structure describes decoded (raw) audio or video data.
Definition frame.h:472
uint8_t * data[AV_NUM_DATA_POINTERS]
pointer to the picture/channel planes.
Definition frame.h:493
int linesize[AV_NUM_DATA_POINTERS]
For video, a positive or negative value, which is typically indicating the size in bytes of each pict...
Definition frame.h:517
AVOption.
Definition opt.h:428
Descriptor that unambiguously describes how the bits of a pixel are stored in the up to 4 data planes...
Definition pixdesc.h:69
int planewidth[4]
Definition vf_dblur.c:39
float radius
Definition vf_dblur.c:33
int planeheight[4]
Definition vf_dblur.c:40
float * buffer
Definition vf_dblur.c:41
int nb_planes
Definition vf_dblur.c:42
float angle
Definition vf_dblur.c:32
#define av_malloc_array(a, b)
#define av_freep(p)
#define src
Definition vp8dsp.c:248
static FILE * out
Definition movenc.c:55
static AVFormatContext * ctx
Definition movenc.c:49
#define height
Definition dsp.h:89
#define width
Definition dsp.h:89
const char * g
Definition vf_curves.c:128
static int config_input(AVFilterLink *inlink)
Definition vf_dblur.c:145
#define f(n, m)
Definition vf_dblur.c:57
static void set_params(DBlurContext *s, float angle, float r)
Definition vf_dblur.c:167
static int filter_frame(AVFilterLink *inlink, AVFrame *in)
Definition vf_dblur.c:191
static int filter_horizontally(AVFilterContext *ctx, int width, int height)
Definition vf_dblur.c:59
static av_cold void uninit(AVFilterContext *ctx)
Definition vf_dblur.c:138
static const AVOption dblur_options[]
Definition vf_dblur.c:48
static const AVFilterPad dblur_inputs[]
Definition vf_dblur.c:293
#define OFFSET(x)
Definition vf_dblur.c:45
static void diriir2d(AVFilterContext *ctx, int plane)
Definition vf_dblur.c:106
static double b1(void *priv, double x, double y)
Definition vf_xfade.c:2034
static double a0(void *priv, double x, double y)
Definition vf_xfade.c:2028
static double a3(void *priv, double x, double y)
Definition vf_xfade.c:2031
static double b0(void *priv, double x, double y)
Definition vf_xfade.c:2033
static double a2(void *priv, double x, double y)
Definition vf_xfade.c:2030
static double a1(void *priv, double x, double y)
Definition vf_xfade.c:2029
const AVFilterPad ff_video_default_filterpad[1]
An AVFilterPad array whose only entry has name "default" and is of type AVMEDIA_TYPE_VIDEO.
Definition video.c:37
AVFrame * ff_get_video_buffer(AVFilterLink *link, int w, int h)
Request a picture buffer with a specific set of permissions.
Definition video.c:89
static double c[64]